import datetime
import logging
import secrets
from typing import List, Dict, Optional, Union, Any, Type, TYPE_CHECKING

import bson
import orjson
from Crypto.Cipher import AES
from bson import ObjectId
from pymongo.results import DeleteResult
from motor.motor_asyncio import AsyncIOMotorDatabase

from alaric import Document, util
from alaric.abc import Buildable, Filterable, Saveable
from alaric.encryption import (
    EncryptedFields,
    HashedFields,
    IgnoreFields,
    AutomaticHashedFields,
)
from alaric.projections import Projection, Show
from alaric.document import T

log = logging.getLogger(__name__)


# noinspection DuplicatedCode
class EncryptedDocument(Document):
    _version = 1

    def __init__(
        self,
        database: AsyncIOMotorDatabase,
        document_name: str,
        *,
        encryption_key: bytes,
        hashed_fields: Optional[HashedFields] = None,
        automatic_hashed_fields: Optional[AutomaticHashedFields] = None,
        encrypted_fields: Optional[EncryptedFields] = None,
        converter: Optional[Type[T]] = None,
        encrypt_all_fields: bool = False,
    ):
        """
        Parameters
        ----------
        database: AsyncIOMotorDatabase
            The database we are connected to
        document_name: str
            What this collection should be called
        encryption_key: bytes
            The key to use for AES encryption
        hashed_fields: Optional[HashedFields]
            A list of fields to SHA512 hash when encountered
        automatic_hashed_fields: Optional[AutomaticHashedFields]
            A list of fields to create an additional column in
            the db for with a hashed variant without exposing
            the hashed data to the end user.
        encrypted_fields: Optional[EncryptedFields]
            A list of fields to AES encrypt when encountered
        converter: Optional[Type[:py:class:`~alaric.document.T`]]
            An optional class to try
            to convert all data-types which
            return either Dict or List into
        encrypt_all_fields: bool
            If set to True, encrypt all fields regardless of
            `hashed_fields` and `encrypted_fields` options.

            This option respects ignored fields.


        .. code-block:: python
            :linenos:

            from motor.motor_asyncio import AsyncIOMotorClient

            client = AsyncIOMotorClient(connection_url)
            database = client["my_database"]
            config_document = Document(database, "config")
        """
        super().__init__(database, document_name, converter=converter)
        self._encryption_key = encryption_key
        self._hashed_fields: HashedFields = (
            hashed_fields if hashed_fields is not None else HashedFields()
        )
        self._automatic_hashed_fields: AutomaticHashedFields = (
            automatic_hashed_fields
            if automatic_hashed_fields is not None
            else AutomaticHashedFields()
        )
        self._encrypted_fields: EncryptedFields = (
            encrypted_fields if encrypted_fields is not None else EncryptedFields()
        )
        self._encrypt_all_fields: bool = encrypt_all_fields

    def __repr__(self):
        return f"<Document(document_name={self._document_name})>"

    @classmethod
    def generate_aes_key(cls) -> bytes:
        """Generate a valid AES key for usage with this class.

        The output should be stored in an environment variable
        for future usage as otherwise you will lose your data.

        For storage purposes, you may find the following methods useful:
         - bytes.hex()
         - bytes.fromhex()

        Returns
        -------
        bytes
            A valid AES key
        """
        # 256 bit
        return secrets.token_bytes(32)

    def _encrypt_data(self, data: Dict, *, ignore_fields: IgnoreFields) -> Dict:
        encrypted_fields = {}
        for k, v in data.items():
            if k in ignore_fields:
                encrypted_fields[k] = v
                continue

            if k in self._automatic_hashed_fields:
                new_key = f"{k}_hashed"
                if new_key in encrypted_fields or new_key in data:
                    raise ValueError(
                        f"Cannot automatically hash {k} as the column {new_key} already exists in the dataset."
                    )

                encrypted_fields[new_key] = util.hash_field(new_key, v)

            if self._encrypt_all_fields:
                v = self._aes_encrypt_field(v)

            elif k in self._encrypted_fields:
                v = self._aes_encrypt_field(v)

            elif k in self._hashed_fields:
                v = util.hash_field(k, v)

            encrypted_fields[k] = v
        return encrypted_fields

    def _decrypt_data(self, data: Dict) -> Dict:
        # Keep the cursor mirror up to date
        decrypted_fields = {}
        # preprocess out automatic hashed fields
        for ktr in [f"{k}_hashed" for k in self._automatic_hashed_fields]:
            data.pop(ktr, None)

        for k, v in data.items():
            if k in self._encrypted_fields or self._encrypt_all_fields:
                try:
                    if not isinstance(v, bson.ObjectId):
                        v = self._aes_decrypt_field(bytes.fromhex(v))
                except ValueError:
                    raise ValueError("Invalid encryption_key in use for this data.")

            decrypted_fields[k] = v

        return decrypted_fields

    def _aes_encrypt_field(self, value) -> Union[str, ObjectId]:
        # Data is stored in the format b'nonce(16 bytes)tag(16 bytes)ciphertext(remaining)'
        cipher = AES.new(self._encryption_key, AES.MODE_GCM)
        if isinstance(value, str):
            value = f"str     |{value}"
        elif isinstance(value, bool):
            value = f"bool    |{1 if value else 0}"
        elif isinstance(value, int):
            value = f"int     |{value}"
        elif isinstance(value, float):
            value = f"float   |{value}"
        elif isinstance(value, datetime.datetime):
            value = f"datetime|{value.isoformat()}"
        elif isinstance(value, list):
            pre = {"list": value}
            value = f"list    |{orjson.dumps(pre).hex()}"
        elif isinstance(value, dict):
            value = f"dict    |{orjson.dumps(value).hex()}"
        elif isinstance(value, ObjectId):
            log.debug("You asked me to encrypt an ObjectId instance, I can't do that.")
            return value
        else:
            raise ValueError(
                f"{value.__class__.__name__} is not yet supported for encryption"
            )

        if not isinstance(value, bytes):
            value = bytes(value, "utf-8")

        ciphertext, tag = cipher.encrypt_and_digest(value)
        ba = bytearray(cipher.nonce)
        ba.extend(tag)
        ba.extend(ciphertext)
        return ba.hex()

    def _aes_decrypt_field(self, value: bytes):
        # Keep the cursor mirror up to date
        # We assume by here it's an AES field
        def extract_list(data) -> List:
            inner = orjson.loads(bytes.fromhex(data))
            return inner["list"]

        def extract_dict(data):
            return orjson.loads(bytes.fromhex(data))

        def extract_bool(data):
            return True if data == "1" else False

        # Whitelist types
        mappings = {
            "str": str,
            "int": int,
            "float": float,
            "bool": extract_bool,
            "dict": extract_dict,
            "list": extract_list,
            "datetime": datetime.datetime.fromisoformat,
        }
        nonce = value[:16]
        tag = value[16:32]
        ciphertext = value[32:]
        cipher = AES.new(self._encryption_key, AES.MODE_GCM, nonce)
        text = cipher.decrypt_and_verify(ciphertext, tag)
        text = text.decode("utf-8")
        _type = text[:9].split("|")[0].strip()
        content = text[9:]
        return mappings[_type](content)

    @staticmethod
    def __ensure_ignore_fields(ignore_fields: Optional[IgnoreFields]):
        if ignore_fields is not None:
            return ignore_fields

        return IgnoreFields()

    def __preprocess_fields(
        self, data: Union[Dict, Saveable], ignore_fields: IgnoreFields
    ):
        if isinstance(data, Saveable):
            data = data.as_dict()

        if not isinstance(data, dict):
            raise ValueError(f"Expected dict, got {data.__class__.__name__}")

        return self._encrypt_data(data, ignore_fields=ignore_fields)

    async def _attempt_convert(
        self, data: Union[Dict, List[Dict]]
    ) -> Union[List[Union[Dict[str, Any], Type[T]]], Union[Dict[str, Any], Type[T]]]:
        if not data:
            return data

        if not self.converter:
            if isinstance(data, list):
                out = []
                for itr in data:
                    out.append(self._decrypt_data(itr))
                return out

            return self._decrypt_data(data)

        if not isinstance(data, list):
            return self.converter(**self._decrypt_data(data))

        new_data = []
        for d in data:
            new_data.append(self.converter(**self._decrypt_data(d)))

        return new_data

    async def insert(
        self,
        data: Union[Dict[str, Any], Saveable],
        *,
        ignore_fields: Optional[IgnoreFields] = None,
    ) -> None:
        """Insert the provided data into the document.

        Parameters
        ----------
        data: Union[Dict[str, Any], Saveable]
            The data to insert
        ignore_fields: Optional[IgnoreFields]
            Any fields to ignore during the hashing / encryption step.

            Useful if your passing this method an already hashed value
            and you don't want to hash the hash.

        .. code-block:: python
            :linenos:

            # If you don't provide an _id,
            # Mongo will generate one for you automatically
            await Document.insert({"_id": 1, "data": "hello world"})
        """
        ignore_fields = self.__ensure_ignore_fields(ignore_fields=ignore_fields)
        data = self.__preprocess_fields(data, ignore_fields=ignore_fields)
        await super().insert(data)

    async def update(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        update_data: Union[Dict[str, Any], Saveable],
        option: str = "set",
        *args: Any,
        ignore_fields: Optional[IgnoreFields] = None,
        **kwargs: Any,
    ) -> None:
        """Performs an UPDATE operation.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
            The data to filter on
        update_data: Union[Dict[str, Any], Saveable]
            The data to upsert
        option: str
            The optional option to pass to mongo,
            default is set

            https://www.mongodb.com/docs/manual/reference/operator/update/
        ignore_fields: Optional[IgnoreFields]
            Any fields to ignore during the hashing / encryption step.

            Useful if your passing this method an already hashed value
            and you don't want to hash the hash.


        .. code-block:: python
            :linenos:

            # Update the document with an `_id` of 1
            # So that it now equals the second argument
            await Document.upsert({"_id": 1}, {"_id": 1, "data": "new data"})
        """
        ignore_fields = self.__ensure_ignore_fields(ignore_fields=ignore_fields)
        update_data = self.__preprocess_fields(update_data, ignore_fields=ignore_fields)
        await super().update(filter_dict, update_data)

    async def upsert(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        update_data: Union[Dict[str, Any], Saveable],
        option: str = "set",
        *args: Any,
        ignore_fields: Optional[IgnoreFields] = None,
        **kwargs: Any,
    ) -> None:
        """Performs an UPSERT operation.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
            The data to filter on
        update_data: Union[Dict[str, Any], Saveable]
            The data to upsert
        option: str
            Update operator.

            https://www.mongodb.com/docs/manual/reference/operator/update/
        ignore_fields: Optional[IgnoreFields]
            Any fields to ignore during the hashing / encryption step.

            Useful if your passing this method an already hashed value
            and you don't want to hash the hash.


        .. code-block:: python
            :linenos:

            # Update the document with an `_id` of `1`
            # So that it now equals the second argument
            # NOTE: If a document with an `_id` of `1`
            # does not exist, then this method will
            # insert the data instead.
            await Document.update({"_id": 1}, {"_id": 1, "data": "new data"})
        """
        ignore_fields = self.__ensure_ignore_fields(ignore_fields=ignore_fields)
        update_data = self.__preprocess_fields(update_data, ignore_fields=ignore_fields)
        await super().upsert(filter_dict, update_data)

    async def increment(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        field: str,
        amount: Union[int, float],
    ) -> None:
        """Increment the provided field.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
            The 'thing' we want to increment
        field: str
            The key for the field to increment
        amount: Union[int, float]
            How much to increment (or decrement) by

        Notes
        -----
        This seamlessly handles incrementing encrypted fields.

        .. code-block:: python
            :linenos:

            # Assuming a data structure of
            # {"_id": 1, "counter": 4}
            await Document.increment({"_id": 1}, "counter", 1)

            # Now looks like
            # {"_id": 1, "counter": 5}

        Raises
        ------
        ValueError
            Nested field updates on encrypted fields is not supported.
        ValueError
            Item to increment didn't exist with this filter.

        Notes
        -----
        You can also use negative numbers to
        decrease the count of a field.
        """
        if field not in self._encrypted_fields:
            return await super().increment(filter_dict, field, amount)

        if "." in field:
            raise ValueError(
                "Nested field updates on encrypted fields is not supported."
            )

        data: Dict = await self.find(  # noqa
            filter_dict, projections=Projection(Show(field)), try_convert=False
        )
        if not data:
            raise ValueError("Item to increment didn't exist with this filter.")

        data[field] += amount
        await self.update(filter_dict, data)

    async def change_field_to(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        field: str,
        new_value: Any,
    ) -> None:
        """Modify a single field and change the value.

        Parameters
        ----------
        filter_dict: Union[Dict[Any, Any], Buildable, Filterable]
            A dictionary to use as a filter or
            :py:class:`AQ` object.
        field: str
            The key for the field to increment
        new_value: Any
            What the field should get changed to


        .. code-block:: python
            :linenos:

            # Assuming a data structure of
            # {"_id": 1, "prefix": "!"}
            await Document.change_field_to({"_id": 1}, "prefix", "?")

            # This will now look like
            # {"_id": 1, "prefix": "?"}
        """
        if field in self._encrypted_fields:
            new_value = self._aes_encrypt_field(new_value)

        await super().change_field_to(filter_dict, field, new_value)

    async def bulk_insert(
        self,
        data: List[Dict],
        ignore_fields: Optional[IgnoreFields] = None,
    ) -> None:
        """
        Given a List of Dictionaries, bulk insert all
        the given dictionaries in a single call.

        Notes
        -----
        Supports encrypted and hashed fields.

        Parameters
        ----------
        data: List[Dict]
            The data to bulk insert
        ignore_fields: Optional[IgnoreFields]
            Any fields to ignore during the hashing / encryption step.

            Useful if your passing this method an already hashed value
            and you don't want to hash the hash.


        .. code-block:: python
            :linenos:

            # Insert 25 documents
            await Document.bulk_insert(
                {"_id": i}
                for i in range(25)
            )
        """
        ignore_fields = self.__ensure_ignore_fields(ignore_fields=ignore_fields)
        self._ensure_list_of_dicts(data)
        encrypted_data = []
        for d in data:
            encrypted_data.append(self._encrypt_data(d, ignore_fields=ignore_fields))

        await self._document.insert_many(encrypted_data)

    if TYPE_CHECKING:
        # I don't like this but Pycharm doesn't
        # properly support methods from parents
        async def find(
            self,
            filter_dict: Union[Dict[str, Any], Buildable, Filterable],
            projections: Optional[Union[Dict[str, Any], Projection]] = None,
            *,
            try_convert: bool = True,
        ) -> Optional[Union[Dict[str, Any], Type[T]]]:
            ...

        async def find_many(
            self,
            filter_dict: Union[Dict[str, Any], Buildable, Filterable],
            projections: Optional[Union[Dict[str, Any], Projection]] = None,
            *,
            try_convert: bool = True,
        ) -> List[Union[Dict[str, Any], Type[T]]]:
            ...

        async def delete(
            self,
            filter_dict: Union[Dict, Buildable, Filterable],
        ) -> Optional[DeleteResult]:
            ...

        async def get_all(
            self,
            filter_dict: Optional[Union[Dict[str, Any], Buildable, Filterable]] = None,
            projections: Optional[Union[Dict[str, Any], Projection]] = None,
            *args: Any,
            try_convert: bool = True,
            **kwargs: Any,
        ) -> List[Optional[Union[Dict[str, Any], Type[T]]]]:
            ...

        async def count(
            self, filter_dict: Union[Dict[Any, Any], Buildable, Filterable]
        ) -> int:
            ...



from __future__ import annotations

from typing import TYPE_CHECKING, List, Dict, Union

if TYPE_CHECKING:
    from alaric.abc import ComparisonT, LogicalT, Buildable


class OR:
    """Conduct a logical OR between all items passed to the constructor.

    Lets build a check using a simple OR which will return
    all results the field ``my_field`` is either ``True``
    OR a number in the list ``[1, 2, 3, 7, 8, 9]``

    .. code-block:: python
        :linenos:

        from alaric import AQ
        from alaric.logical import OR
        from alaric.comparison import EQ, IN

        query = AQ(OR(EQ("my_field", True), IN("my_field", [1, 2, 3, 7, 8, 9])))
    """

    def __init__(
        self,
        *comparisons: Union[
            Union[ComparisonT, LogicalT, Buildable],
            List[Union[ComparisonT, LogicalT, Buildable]],
        ],
    ):
        self.comparisons: List[Union[ComparisonT, LogicalT, Buildable]] = list(
            comparisons
        )

    def __repr__(self):
        return f"OR(comparisons={self.comparisons})"

    def build(self) -> Dict[str, List[Dict]]:
        """Return this instance as a usable Mongo filter."""
        comparisons: List[Dict] = [c.build() for c in self.comparisons]
        return {"$or": comparisons}



from typing import Dict


class Regex:
    """
    Provides an interface for fetching data based on a given regex.

    Let's match all documents where the field ``title``
    contains the word "Blade"

    .. code-block:: python

        from alaric import AQ
        from alaric.comparison import Regex

        query = AQ(Regex("title", "Blade"))

    For further options see:
    https://www.mongodb.com/docs/manual/reference/operator/query/regex/
    """

    def __init__(
        self,
        field: str,
        regex: str,
        *,
        case_insensitive: bool = False,
        other_options: str = "",
    ):
        self.field: str = field
        self.regex = regex
        self.case_insensitive = case_insensitive
        self.options = other_options
        if case_insensitive:
            self.options += "i"

    def __repr__(self):
        return f"Regex({self.regex=}, {self.options=})"

    def build(self) -> Dict[str, Dict[str, str]]:
        """Return this instance as a usable Mongo filter."""
        return {self.field: {"$regex": self.regex, "$options": self.options}}



from .base import Base
from .hashed_fields import HashedFields
from .encrypted_fields import EncryptedFields
from .ignore_fields import IgnoreFields
from .hashed_query_field import HashedQueryField
from .automatic_hashed_fields import AutomaticHashedFields

HQF = HashedQueryField



from alaric.encryption import Base


class AutomaticHashedFields(Base):
    """A list of fields which should have a hashed field
    created automatically in the background but without access to the user.

    I.e. Alaric when told to automatically hash the field ``guild_id``
    will look at incoming data, create an extra field called ``guild_id_hashed``
    and add it to the database. Whenever you fetch data from the database,
    Alaric will remove this field so you never see it.

    If you want to use this field in your code,
    use :py:class:`alaric.encryption.HashedFields` instead.

    .. code-block:: python
        :linenos:

        from alaric.encryption import AutomaticHashedFields

        AutomaticHashedFields("guild_id", "test")

    .. note::
        You can use this in conjunction with an encrypted field to
        have a hash representing the unencrypted data.
    """



from typing import Dict, Union


class IN:
    """
    Asserts the provided field is within the provided values.

    This class can be used in conjunction with :py:class:`alaric.meta.NEGATE`

    Parameters
    ----------
    field: str
        The field to check in.
    value: Union[list, tuple, set]
        A iterable of values that field should be in.


    Lets match all documents where the field ``counter``
    is equal to any number in this list.

    .. code-block:: python

        from alaric import AQ
        from alaric.comparison import IN

        query = AQ(IN("counter", [1, 3, 5, 7, 9]))
    """

    def __init__(self, field: str, value: Union[list, tuple, set]):
        self.field: str = field
        assert isinstance(value, (list, tuple, set))
        self.value: Union[list, tuple, set] = value

        self._operator = "$in"

    def __repr__(self):
        return f"IN(field='{self.field}', value={self.value})"

    def build(self) -> Dict[str, Dict[str, Union[list, tuple, set]]]:
        """Return this instance as a usable Mongo filter."""
        return {self.field: {self._operator: self.value}}



from typing import runtime_checkable, Protocol, Dict


@runtime_checkable
class Buildable(Protocol):
    """Protocol for class based Queries."""

    def build(self) -> Dict:
        """Return this instance as a usable Mongo filter."""
        ...



from __future__ import annotations

import datetime
from typing import (
    TYPE_CHECKING,
    Optional,
    Any,
    Dict,
    List,
    Tuple,
    Union,
    Literal,
    Type,
    TypeVar,
)

from motor.motor_asyncio import AsyncIOMotorCollection, AsyncIOMotorCursor

from alaric.abc import Buildable, Filterable
from alaric.encryption import EncryptedFields, AutomaticHashedFields
from alaric.meta import All
from alaric.projections import Projection

if TYPE_CHECKING:
    from alaric import Document

C = TypeVar("C")
"""A typevar representing the type of a given converter class"""


# noinspection DuplicatedCode
class Cursor:
    def __init__(
        self,
        collection: AsyncIOMotorCollection,
        *,
        converter: Optional[Type[C]] = None,
        encryption_key: Optional[bytes] = None,
        encrypted_fields: Optional[EncryptedFields] = None,
        automatic_hashed_fields: Optional[AutomaticHashedFields] = None,
    ):
        """

        Parameters
        ----------
        collection: AsyncIOMotorCollection
            The motor collection
        converter: Optional[Type[:py:class:`~alaric.cursor.C`]]
            An optional class to try
            to convert all data-types which
            return either Dict or List into
        encrypted_fields: Optional[EncryptedFields]
            A list of fields to AES decrypt when encountered
        encryption_key: Optional[bytes]
            The key to use for AES decryption
        automatic_hashed_fields: Optional[AutomaticHashedFields]
            A list of fields to create an additional column in
            the db for with a hashed variant without exposing
            the hashed data to the end user.

        Notes
        -----
        This class is iterable using ``async for``
        """
        self._collection: AsyncIOMotorCollection = collection
        self._filter: Dict[str, Any] = {}
        self._projections: Optional[Dict[str, Any]] = None
        self._limit: int = 0  # Amount of items to return
        self._sort: Optional[List[Tuple[str, Any]], Tuple[str, Any]] = None
        self._cursor: Optional[AsyncIOMotorCursor] = None
        self._converter: Optional[Type[C]] = converter

        if encrypted_fields and not encryption_key:
            raise ValueError(
                "You must set both encryption options to use encryption features."
            )

        self._encrypted_fields: EncryptedFields = (
            encrypted_fields if encrypted_fields is not None else EncryptedFields()
        )
        self._automatic_hashed_fields: AutomaticHashedFields = (
            automatic_hashed_fields
            if automatic_hashed_fields is not None
            else AutomaticHashedFields()
        )
        self._encryption_key = encryption_key

    @classmethod
    def from_document(cls, document: Document) -> Cursor:
        """Create a new :py:class:`~alaric.Cursor` from a :py:class:`~alaric.Document`"""
        return cls(document.raw_collection, converter=document.converter)

    @staticmethod
    def __ensure_built(data: Union[Dict, Buildable, Filterable]) -> Dict:
        if isinstance(data, Filterable):
            return data.as_filter()

        elif isinstance(data, Buildable):
            return data.build()

        return data

    def _ensure_modifiable(self):
        if self._cursor is not None:
            raise ValueError(
                "Cursor is already in use, create a copy and modify that instance."
            )

    def copy(self) -> Cursor:
        """Returns a modifiable version of this cursor.

        Returns
        -------
        Cursor
            A new cursor instance
        """
        cursor: Cursor = Cursor(self._collection)
        cursor._sort = self._sort
        cursor._limit = self._limit
        cursor._filter = self._filter
        cursor._projections = self._projections
        return cursor

    def _build_cursor(self):
        self._ensure_modifiable()
        if self._projections:
            motor_cursor = self._collection.find(self._filter, self._projections)
        else:
            motor_cursor = self._collection.find(self._filter)

        if self._sort:
            motor_cursor = motor_cursor.sort(self._sort)
        if self._limit:
            motor_cursor = motor_cursor.limit(self._limit)

        self._cursor = motor_cursor

    def set_filter(
        self,
        filter_data: Union[
            Dict[str, Any],
            Buildable,
            Filterable,
        ] = All(),
    ) -> Cursor:
        """Set the filter for the cursor query.

        Parameters
        ----------
        filter_data: Union[Dict[str, Any], Buildable, Filterable]
            A dictionary to use as a filter or
            :py:class:`AQ` object.

        Returns
        -------
        Cursor
            This cursor instance for method chaining.
        """
        self._ensure_modifiable()
        self._filter = self.__ensure_built(filter_data)
        return self

    def set_projections(
        self,
        projections: Optional[Union[Dict[str, Literal[0, 1]], Projection]] = None,
    ) -> Cursor:
        """Define what data should be returned for each document in the result

        Parameters
        ----------
        projections: Optional[Union[Dict[str, Literal[0, 1]], Projection]]
            Specify the data you want returned from matching queries.

        Returns
        -------
        Cursor
            This cursor instance for method chaining.
        """
        self._ensure_modifiable()
        self._projections = self.__ensure_built(projections)
        return self

    def set_limit(self, limit: int = 0) -> Cursor:
        """Set a limit for how many documents should
        be returned in the query.

        Use ``0`` to indicate no limit.

        Parameters
        ----------
        limit: int
            How many documents should be returned.

            Defaults to no limit.

        Returns
        -------
        Cursor
            This cursor instance for method chaining.

        Raises
        ------
        ValueError
            You specified a negative number.
        """
        self._ensure_modifiable()
        if not isinstance(limit, int) or limit < 0:
            raise ValueError("Positive numbers only")

        self._limit = limit
        return self

    def set_sort(
        self, order: Optional[List[Tuple[str, Any]], Tuple[str, Any]] = None
    ) -> Cursor:
        """Set the sort order for returned results

        Parameters
        ----------
        order: Optional[List[Tuple[str, Any]], Tuple[str, Any]]
            The order to sort by

        Returns
        -------
        Cursor
            This cursor instance for method chaining

        Raises
        ------
        ValueError
            The passed value was not a list or tuple

        .. code-block:: python
            :linenos:

            import alaric

            # Lets sort by the count field
            Cursor.set_sort(("count", alaric.Ascending))

            ...
            # Lets sort first by the count field,
            # then also by the backup_count field
            Cursor.set_sort([("count", alaric.Ascending), ("backup_count", alaric.Descending)])
        """
        self._ensure_modifiable()
        # TODO: Make a sort object with fields to more easily support advanced sorting?
        if not isinstance(order, list) and order is not None:
            if isinstance(order, tuple):
                order = [order]
            else:
                raise ValueError("order must be either a tuple, or list of tuples")

        self._sort = order
        return self

    async def execute(self) -> List[Union[Dict[str, Any], Type[C]]]:
        """Execute this cursor and return the result."""
        self._build_cursor()
        data = await self._cursor.to_list(None)
        return await self._try_convert(data)

    def __aiter__(self):
        """
        This style of iteration is also supported.

        .. code-block:: python
            :linenos:

            cursor: Cursor = ...
            async for entry in cursor:
                print(entry)

        """
        self._build_cursor()
        return self

    async def __anext__(self):
        async for data in self._cursor:
            return await self._try_convert(data)

        raise StopAsyncIteration

    async def _try_convert(
        self, data: Union[Dict, List[Dict]]
    ) -> Union[List[Union[Dict[str, Any], Type[C]]], Union[Dict[str, Any], Type[C]]]:
        if not data:
            return data

        if not self._converter:
            if isinstance(data, list):
                out = []
                for itr in data:
                    out.append(self._decrypt_data(itr))
                return out

            return self._decrypt_data(data)

        if not isinstance(data, list):
            return self._converter(**self._decrypt_data(data))

        new_data = []
        for d in data:
            new_data.append(self._converter(**self._decrypt_data(d)))

        return new_data

    # Copied from EncryptedDocument
    def _decrypt_data(self, data: Dict) -> Dict:
        decrypted_fields = {}
        for ktr in [f"{k}_hashed" for k in self._automatic_hashed_fields]:
            data.pop(ktr, None)

        for k, v in data.items():
            if k in self._encrypted_fields:
                try:
                    v = self._aes_decrypt_field(bytes.fromhex(v))
                except ValueError:
                    raise ValueError("Invalid encryption_key in use for this data.")

            decrypted_fields[k] = v

        return decrypted_fields

    def _aes_decrypt_field(self, value: bytes):
        def extract_list(data) -> List:
            import orjson

            inner = orjson.loads(bytes.fromhex(data))
            return inner["list"]

        def extract_dict(data):
            import orjson

            return orjson.loads(bytes.fromhex(data))

        def extract_bool(data):
            return True if data == "1" else False

        # Whitelist types
        mappings = {
            "str": str,
            "int": int,
            "float": float,
            "bool": extract_bool,
            "dict": extract_dict,
            "list": extract_list,
            "datetime": datetime.datetime.fromisoformat,
        }
        nonce = value[:16]
        tag = value[16:32]
        ciphertext = value[32:]
        from Crypto.Cipher import AES

        cipher = AES.new(self._encryption_key, AES.MODE_GCM, nonce)
        text = cipher.decrypt_and_verify(ciphertext, tag)
        text = text.decode("utf-8")
        _type = text[:9].split("|")[0].strip()
        content = text[9:]
        return mappings[_type](content)



from __future__ import annotations

import io
import logging
from datetime import timedelta
from typing import (
    TYPE_CHECKING,
    List,
    Union,
    Dict,
    Any,
    Optional,
    TypeVar,
    cast,
    Generic,
)

import orjson

from alaric.abc import Buildable, Filterable, Saveable

if TYPE_CHECKING:
    from redis.asyncio.client import Redis
    from alaric import Document


log = logging.getLogger(__name__)
C = TypeVar("C")
"""A typevar representing the type of a given converter class"""


class CachedDocument(Generic[C]):
    """This document implements a cache in front of MongoDB for read heavy work flows.

    Read process:

    .. code-block:: text

        result = attempt to hit redis
        if result is None:
            result = fetch from database
            populate redis cache

    Write process:

    .. code-block:: text

        Push changes back to database
        update redis cache

    Notes
    -----
    This document works off the assumption that a documents ``_id`` remains
    consistent through the lifetime of the entry.

    This document will also leave hanging redis entries when lookups
    outside the _id are modified during the lifetime of the object.
    This is mitigated by enforcing a TTL of all redis entries.
    """

    def __init__(
        self,
        *,
        document: Document,
        redis_client: Redis,
        extra_lookups: List[List[str]] = None,
        cache_ttl: timedelta = timedelta(hours=1),
    ):
        """

        Parameters
        ----------
        document: alaric.Document
            The underlying DB document
        redis_client: redis.asyncio.client.Redis
            The Redis instance to use
        extra_lookups: List[List[str]]
            Extra lookups to build. For example:

            .. code-block:: python

                class Test:
                    def __init__(self, _id, data):
                        self.data = data
                        self._id = _id

                ...

                extra_lookups = [["data"]]

            Would let you use ``get`` with either the
            ``_id`` or ``data`` parameter of ``Test``

        cache_ttl: timedelta
            How long keys should exist in Redis.

            This is a requirement as this class will
            leave hanging keys in Redis when certain values change.
        """
        self.document: Document = document
        self._redis_client: Redis = redis_client
        self._cache_ttl: timedelta = cache_ttl
        self._extra_lookups: List[List[str]] = []
        if extra_lookups is not None:
            # keys are sorted so that we can consistently
            # do redis lookups without the need for more than
            # a straight string key lookup
            for lookup in extra_lookups:
                self._extra_lookups.append(sorted(lookup))

    async def get(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        *,
        try_convert: bool = True,
    ) -> Optional[Union[Dict[str, Any], C]]:
        """Fetch a document.

        Attempts to fetch from Redis and then falls back to DB

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
        try_convert: bool
            See :py:class:`alaric.Document`

        Returns
        -------
        Optional[Union[Dict[str, Any], C]]
            The data fetched from either Redis or your DB

        Notes
        -----
        Due to how items are stored internally,
        filter_dict must eval to a literal str: str
        pairing otherwise this will fall back to the DB

        """
        filter_dict: Dict[str, str] = self.document._ensure_built(filter_dict)
        lookup_key = original_key = self._build_redis_lookup_key(filter_dict)

        # If not a straight _id lookup, resolve the chain
        # back to the raw data itself. We also assume that
        # any lookup key which does not start with _id
        # requires resolving back the source
        if not lookup_key.startswith("_id:"):
            new_lookup_key = await self._redis_client.get(lookup_key)
            if new_lookup_key is not None:
                # If It's None, just do the check as is and
                # call it a day
                lookup_key = new_lookup_key

        result = await self._redis_client.get(lookup_key)
        if result is None:
            result = await self.document.find(filter_dict, try_convert=False)
            if result is not None:
                result = cast(Dict[str, Any], result)
                await self._update_redis_cache(result)

            log.debug("Cache miss for %s", original_key)
        else:
            result = orjson.loads(result)
            log.debug("Cache hit for %s", original_key)

        if try_convert:
            return await self.document._attempt_convert(result)
        return result

    async def set(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        update_data: Union[Dict[str, Any], Saveable],
    ) -> None:
        """Write to Redis and the DB.

        Notes
        -----
        This performs an UPSERT operation
        on the database.

        This also requires ``_id`` to be set on update_data
        in order to cache this to Redis

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
        update_data: Union[Dict[str, Any], Saveable]
        """
        filter_dict = self.document._ensure_built(filter_dict)
        update_data = self.document._ensure_insertable(update_data)
        if "_id" not in update_data:
            log.warning("Failed to cache data as _id was missing: %s", update_data)
        else:
            await self._update_redis_cache(update_data)

        await self.document.upsert(filter_dict, update_data)

    async def _update_redis_cache(self, data: Dict[str, Any]):
        """Updates the redis cache data entries"""
        assert "_id" in data
        data_id = data["_id"]
        data_id_key = f"_id:{data_id}|"
        data_str = orjson.dumps(data)
        await self._redis_client.setex(data_id_key, self._cache_ttl, data_str)

        for lookup_entry in self._extra_lookups:
            key = io.StringIO()
            for item in lookup_entry:
                key.write(f"{item}:{data[item]}|")

            await self._redis_client.setex(key.getvalue(), self._cache_ttl, data_id_key)

    @staticmethod
    def _build_redis_lookup_key(filter_dict: Dict[str, str]) -> str:
        """Given a dict, build the redis lookup key"""
        # Redis key format
        # field_name:field_value|
        # The key should contain a trailing pipe
        result = io.StringIO()
        for key in sorted(filter_dict.keys()):
            result.write(f"{key}:{filter_dict[key]}|")

        return result.getvalue()



from .show import Show
from .hide import Hide
from .projections import Projection

SHOW = Show
HIDE = Hide
PROJECTION = Projection
__all__ = (
    "Show",
    "Hide",
    "Projection",
    "SHOW",
    "HIDE",
    "PROJECTION",
)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from alaric.abc import ComparisonT


class CI:
    """
    Marks a given option as case-insensitive.
    """

    pass



import hashlib


def hash_field(field, value):
    """Hash a field with SHA512

    Parameters
    ----------
    field: str
        The name of the field your hashing
    value
        The value to hash

    Raises
    ------
    ValueError
        Unsupported type to hash
    """
    if isinstance(value, (int, float, bool)):
        # Support hashing ints, floats and bools
        # for search filters
        value = str(value)

    try:
        return hashlib.sha512(value.encode("utf-8")).hexdigest()
    except TypeError:
        raise ValueError(
            f"Cannot hash field '{field}' as it is an "
            f"unsupported type {value.__class__.__name__}"
        )



import logging
from collections import namedtuple

from .advanced_query import AQ
from .document import Document
from .cursor import Cursor
from .encrypted_document import EncryptedDocument
from .cached_document import CachedDocument

Ascending = 1
Descending = -1

__all__ = (
    "AQ",
    "Document",
    "Ascending",
    "Descending",
    "Cursor",
    "EncryptedDocument",
    "util",
    "CachedDocument",
)
__version__ = "1.5.0"
VersionInfo = namedtuple("VersionInfo", "major minor micro releaselevel serial")
version_info = VersionInfo(major=1, minor=5, micro=0, releaselevel="final", serial=0)
logging.getLogger(__name__).addHandler(logging.NullHandler())



import asyncio
import os
import secrets
from copy import copy
from pprint import pprint
from typing import Dict

from motor.motor_asyncio import AsyncIOMotorClient

from alaric import *
from alaric.encryption import *
from alaric.comparison import *
from alaric.logical import *
from alaric.meta import *
from alaric.projections import *
from alaric.types import *


async def main():
    # client = AsyncIOMotorClient(os.environ["MONGO"])
    # db = client["test_encryption"]

    Negate(Exists("Test"))

    print(AQ(EQ("channel_id", None)).build())


# {'$and': [{'suggestion': {'$regex': 'include these'}, {'state': {'$eq': 'pending'}}]}

if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import logging
from typing import Dict


log = logging.getLogger(__name__)


class All:
    """
    Return all documents in the collection.

    .. code-block:: python
        :linenos:

        from alaric.meta import All

        query = AQ(All())
    """

    def __repr__(self):
        return f"ALL()"

    def build(self) -> Dict:
        return {}



from __future__ import annotations

from typing import List, Dict, Optional, Union, Any, TypeVar, Type, TYPE_CHECKING

from pymongo.results import DeleteResult
from motor.motor_asyncio import AsyncIOMotorDatabase, AsyncIOMotorCollection

from alaric.abc import Buildable, Filterable, Saveable
from alaric.projections import Projection

if TYPE_CHECKING:
    from alaric import Cursor

T = TypeVar("T")
"""A typevar representing the type of a given converter class"""


class Document:
    _version = 11

    def __init__(
        self,
        database: AsyncIOMotorDatabase,
        document_name: str,
        converter: Optional[Type[T]] = None,
    ):
        """
        Parameters
        ----------
        database: AsyncIOMotorDatabase
            The database we are connected to
        document_name: str
            What this _document should be called
        converter: Optional[Type[:py:class:`~alaric.document.T`]]
            An optional class to try
            to convert all data-types which
            return either Dict or List into


        .. code-block:: python
            :linenos:

            from motor.motor_asyncio import AsyncIOMotorClient

            client = AsyncIOMotorClient(connection_url)
            database = client["my_database"]
            config_document = Document(database, "config")
        """
        self._document_name: str = document_name
        self._database: AsyncIOMotorDatabase = database
        self._document: AsyncIOMotorCollection = database[document_name]  # type: ignore

        self.converter: Type[T] = converter

    def __repr__(self):
        return f"<Document(document_name={self.collection_name})>"

    async def find(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        projections: Optional[Union[Dict[str, Any], Projection]] = None,
        *,
        try_convert: bool = True,
    ) -> Optional[Union[Dict[str, Any], Type[T]]]:
        """Find and return one item.

        Parameters
        ----------
        filter_dict: Union[Dict, Buildable, Filterable]
            A dictionary to use as a filter or
            :py:class:`AQ` object.
        projections: Optional[Union[Dict[str, Any], Projection]]
            Specify the data you want
            returned from matching queries.
        try_convert: bool
            Whether to attempt to
            run convertors on returned data.

            Defaults to True

        Returns
        -------
        Optional[Union[Dict[str, Any], Type[:py:class:`~alaric.document.T`]]]
            The result of the query


        .. code-block:: python
            :linenos:

            # Find the document where the `_id` field is equal to `my_id`
            data: dict = await Document.find({"_id": "my_id"})
        """
        filter_dict = self._ensure_built(filter_dict)
        projections = projections or {}
        projections = self._ensure_built(projections)

        if projections:
            data = await self._document.find_one(filter_dict, projections)
        else:
            data = await self._document.find_one(filter_dict)

        if try_convert:
            return await self._attempt_convert(data)
        return data

    async def find_many(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        projections: Optional[Union[Dict[str, Any], Projection]] = None,
        *,
        try_convert: bool = True,
    ) -> List[Union[Dict[str, Any], Type[T]]]:
        """
        Find and return all items
        matching the given filter.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
            A dictionary to use as a filter or
            :py:class:`AQ` object.
        projections: Optional[Union[Dict[str, Any], Projection]]
            Specify the data you want
            returned from matching queries.
        try_convert: bool
            Whether to attempt to
            run convertors on returned data.

            Defaults to True

        Returns
        -------
        List[Union[Dict[str, Any], Type[:py:class:`~alaric.document.T`]]]
            The result of the query

        Notes
        -----
        This uses a cursor internally, consider using them for
        more complicated queries.


        .. code-block:: python
            :linenos:

            # Find all documents where the key `my_field` is `true`
            data: list[dict] = await Document.find_many({"my_field": True})
        """
        filter_dict = self._ensure_built(filter_dict)
        projections = projections or {}
        projections = self._ensure_built(projections)

        if projections:
            data = await self._document.find(filter_dict, projections).to_list(None)  # type: ignore
        else:
            data = await self._document.find(filter_dict).to_list(None)  # type: ignore

        if try_convert:
            return await self._attempt_convert(data)
        return data

    async def delete(
        self,
        filter_dict: Union[Dict, Buildable, Filterable],
    ) -> Optional[DeleteResult]:
        """
        Delete an item from the Document
        if an item with the provided filter exists.

        Parameters
        ----------
        filter_dict: Union[Dict, Buildable, Filterable]
            A dictionary to use as a filter or
            :py:class:`AQ` object.

        Returns
        -------
        Optional[DeleteResult]
            The result of deletion if it occurred.


        .. code-block:: python
            :linenos:

            # Delete items with a `prefix` of `!`
            await Document.delete({"prefix": "!"})
        """
        filter_dict = self._ensure_built(filter_dict)
        result: DeleteResult = await self._document.delete_many(filter_dict)
        result: Optional[DeleteResult] = result if result.deleted_count != 0 else None
        return result

    async def delete_all(self) -> None:
        """Delete all data associated with this document.

        Notes
        -----
        This will attempt to complete the operation
        in a single call, however, if that fails it
        will fall back to manually deleting items one by one.

        Warnings
        --------
        There is no going back if you call this accidentally.

        This also currently doesn't appear to work.
        """
        try:
            await self._document.drop()
        except:  # noqa
            for entry in await self.get_all(try_convert=False):
                await self.delete(entry)

    async def get_all(
        self,
        filter_dict: Optional[Union[Dict[str, Any], Buildable, Filterable]] = None,
        projections: Optional[Union[Dict[str, Any], Projection]] = None,
        *args: Any,
        try_convert: bool = True,
        **kwargs: Any,
    ) -> List[Optional[Union[Dict[str, Any], Type[T]]]]:
        """
        Fetches and returns all items
        which match the given filter.

        Parameters
        ----------
        filter_dict: Optional[Union[Dict[str, Any], Buildable, Filterable]]
            A dictionary to use as a filter or
            :py:class:`AQ` object.
        projections: Optional[Union[Dict[str, Any], Projection]]
            Specify the data you want
            returned from matching queries.
        try_convert: bool
            Whether to attempt to
            run convertors on returned data.

            Defaults to True

        Returns
        -------
        List[Optional[Union[Dict[str, Any], Type[:py:class:`~alaric.document.T`]]]]
            The items matching the filter


        .. code-block:: python
            :linenos:

            data: list[dict] = await Document.get_all()
        """
        filter_dict = filter_dict or {}
        filter_dict = self._ensure_built(filter_dict)

        projections = projections or {}
        projections = self._ensure_built(projections)

        if projections:
            data = await self._document.find(
                filter_dict, projections, *args, **kwargs
            ).to_list(
                None  # type: ignore
            )
        else:
            data = await self._document.find(filter_dict, *args, **kwargs).to_list(None)  # type: ignore

        if try_convert:
            return await self._attempt_convert(data)
        return data

    async def insert(self, data: Union[Dict[str, Any], Saveable]) -> None:
        """Insert the provided data into the document.

        Parameters
        ----------
        data: Union[Dict[str, Any], Saveable]
            The data to insert


        .. code-block:: python
            :linenos:

            # If you don't provide an _id,
            # Mongo will generate one for you automatically
            await Document.insert({"_id": 1, "data": "hello world"})
        """
        data = self._ensure_insertable(data)
        await self._document.insert_one(data)

    async def update(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        update_data: Union[Dict[str, Any], Saveable],
        option: str = "set",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Performs an UPDATE operation.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
            The data to filter on
        update_data: Union[Dict[str, Any], Saveable]
            The data to upsert
        option: str
            The optional option to pass to mongo,
            default is set

            https://www.mongodb.com/docs/manual/reference/operator/update/


        .. code-block:: python
            :linenos:

            # Update the document with an `_id` of 1
            # So that it now equals the second argument
            await Document.update({"_id": 1}, {"_id": 1, "data": "new data"})
        """
        filter_dict = self._ensure_built(filter_dict)
        update_data = self._ensure_insertable(update_data)

        await self._document.update_one(
            filter_dict, {f"${option}": update_data}, *args, **kwargs
        )

    async def upsert(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        update_data: Union[Dict[str, Any], Saveable],
        option: str = "set",
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Performs an UPSERT operation.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
            The data to filter on
        update_data: Union[Dict[str, Any], Saveable]
            The data to upsert
        option: str
            Update operator.

            https://www.mongodb.com/docs/manual/reference/operator/update/


        .. code-block:: python
            :linenos:

            # Update the document with an `_id` of `1`
            # So that it now equals the second argument
            # NOTE: If a document with an `_id` of `1`
            # does not exist, then this method will
            # insert the data instead.
            await Document.upsert({"_id": 1}, {"_id": 1, "data": "new data"})
        """
        filter_dict = self._ensure_built(filter_dict)
        update_data = self._ensure_insertable(update_data)
        await self.update(
            filter_dict, update_data, option, upsert=True, *args, **kwargs
        )

    async def unset(
        self, filter_dict: Union[Dict[str, Any], Buildable, Filterable], field: Any
    ) -> None:
        """Delete a given field on a collection

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
            The fields to match on (Think _id)
        field: Any
            The field to remove


        .. code-block:: python
            :linenos:

            # Assuming we have a document that looks like
            # {"_id": 1, "field_one": True, "field_two": False}
            await Document.unset({"_id": 1}, "field_two")

            # This data will now look like the following
            # {"_id": 1, "field_one": True}
        """
        filter_dict = self._ensure_built(filter_dict)
        await self._document.update_one(filter_dict, {"$unset": {field: True}})

    async def increment(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        field: str,
        amount: Union[int, float],
    ) -> None:
        """Increment the provided field.

        Parameters
        ----------
        filter_dict: Union[Dict[str, Any], Buildable, Filterable]
            The 'thing' we want to increment
        field: str
            The key for the field to increment
        amount: Union[int, float]
            How much to increment (or decrement) by


        .. code-block:: python
            :linenos:

            # Assuming a data structure of
            # {"_id": 1, "counter": 4}
            await Document.increment({"_id": 1}, "counter", 1)

            # Now looks like
            # {"_id": 1, "counter": 5}

        Notes
        -----
        You can also use negative numbers to
        decrease the count of a field.
        """
        filter_dict = self._ensure_built(filter_dict)
        await self._document.update_one(filter_dict, {"$inc": {field: amount}})

    async def change_field_to(
        self,
        filter_dict: Union[Dict[str, Any], Buildable, Filterable],
        field: str,
        new_value: Any,
    ) -> None:
        """Modify a single field and change the value.

        Parameters
        ----------
        filter_dict: Union[Dict[Any, Any], Buildable, Filterable]
            A dictionary to use as a filter or
            :py:class:`AQ` object.
        field: str
            The key for the field to increment
        new_value: Any
            What the field should get changed to


        .. code-block:: python
            :linenos:

            # Assuming a data structure of
            # {"_id": 1, "prefix": "!"}
            await Document.change_field_to({"_id": 1}, "prefix", "?")

            # This will now look like
            # {"_id": 1, "prefix": "?"}
        """
        filter_dict = self._ensure_built(filter_dict)
        await self._document.update_one(filter_dict, {"$set": {field: new_value}})

    async def count(
        self, filter_dict: Union[Dict[Any, Any], Buildable, Filterable]
    ) -> int:
        """Return a count of how many items match the filter.

        Parameters
        ----------
        filter_dict:  Union[Dict[Any, Any], Buildable, Filterable]
            The count filer.

        Returns
        -------
        int
            How many items matched the filter.


        .. code-block:: python
            :linenos:

            # How many items have the `enabled` field set to True
            count: int = await Document.count({"enabled": True})
        """
        filter_dict = self._ensure_built(filter_dict)
        return await self._document.count_documents(filter_dict)

    async def bulk_insert(self, data: List[Dict]) -> None:
        """
        Given a List of Dictionaries, bulk insert all
        the given dictionaries in a single call.

        Parameters
        ----------
        data: List[Dict]
            The data to bulk insert


        .. code-block:: python
            :linenos:

            # Insert 25 documents
            await Document.bulk_insert(
                {"_id": i}
                for i in range(25)
            )
        """
        self._ensure_list_of_dicts(data)
        await self._document.insert_many(data)

    # <-- Private methods -->
    @staticmethod
    def _ensure_list_of_dicts(data: List[Dict]):
        assert isinstance(data, list)
        assert all(isinstance(entry, dict) for entry in data)

    @staticmethod
    def _ensure_insertable(data: Union[Dict, Saveable]) -> Dict:
        if isinstance(data, Saveable):
            return data.as_dict()

        if not isinstance(data, dict):
            raise ValueError(f"Expected dict, got {data.__class__.__name__}")

        return data

    @staticmethod
    def _ensure_built(data: Union[Dict, Buildable, Filterable]) -> Dict:
        if isinstance(data, Filterable):
            return data.as_filter()

        elif isinstance(data, Buildable):
            return data.build()

        return data

    async def _attempt_convert(
        self, data: Union[Dict, List[Dict]]
    ) -> Union[List[Union[Dict[str, Any], Type[T]]], Union[Dict[str, Any], Type[T]]]:
        # This exists purely so we don't lose intelli-sense
        # as decorators remove intelli-sense
        if not data or not self.converter:
            return data

        if not isinstance(data, list):
            return self.converter(**data)

        new_data = []
        for d in data:
            new_data.append(self.converter(**d))

        return new_data

    # <-- Some basic internals -->
    @property
    def collection_name(self) -> str:
        """The connected collections name."""
        return self._document_name

    @property
    def document_name(self) -> str:
        """Same as :py:meth:`~alaric.Document.collection_name`"""
        return self._document_name

    @property
    def raw_database(self) -> AsyncIOMotorDatabase:
        """Access to the database instance."""
        return self._database

    @property
    def raw_collection(self) -> AsyncIOMotorCollection:
        """The connection collection instance."""
        return self._document

    def create_cursor(self) -> Cursor:
        from alaric import Cursor

        return Cursor(self.raw_collection, converter=self.converter)



from .comparison import ComparisonT
from .logical import LogicalT
from .buildable import Buildable
from .filterable import Filterable
from .saveable import Saveable

__all__ = ("ComparisonT", "LogicalT", "Buildable", "Filterable", "Saveable")



from alaric.encryption import Base


class EncryptedFields(Base):
    """A list of fields which should be encrypted when encountered.

    .. code-block:: python
        :linenos:

        from alaric.encryption import EncryptedFields

        EncryptedFields("guild_id", "test")
    """



from typing import Set


class Base:
    def __init__(self, *fields: str):
        self.fields: Set = set(fields)

    def __contains__(self, item):
        return item in self.fields

    def __repr__(self):
        return f"{self.__class__.__name__}({','.join(self.fields)})"

    def __iter__(self):
        return iter(self.fields)



from __future__ import annotations

from typing import Dict, List, Union, cast

from alaric.abc import Buildable
from alaric.projections import Show, Hide


class Projection:
    """Specify that only the given fields should be returned from the query.

    .. code-block:: python
        :linenos:

        # Assuming the data structure
        # {"_id": 1234, "prefix": "!", "has_premium": False}
        data: dict = await Document.find(
            {"_id": "my_id"}, projections=Projection(Show("prefix"))
        )
        print(data)
        # Will print {"prefix": "!"}
    """

    def __init__(self, *fields: Union[Show, Hide]):
        self.fields: List[Union[Show, Hide]] = list(fields)

    def __repr__(self):
        return f"Projection({self.fields})"

    def build(self) -> Dict:
        fields: Dict = {}
        for field in self.fields:
            field = cast(Buildable, field)
            fields = {**fields, **field.build()}

        if "_id" not in fields:
            # Hide unless asked
            fields["_id"] = 0

        return fields



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Dict

from alaric.comparison import IN, GT, LT, EQ
from alaric.comparison.comparison_exists import Exists

if TYPE_CHECKING:
    from alaric.abc import ComparisonT

log = logging.getLogger(__name__)


class Negate:
    """
    Negate a given option, I.e. Do the opposite.

    Supported operands:

    * :py:class:`~alaric.comparison.Exists`
    * :py:class:`~alaric.comparison.IN`
    * :py:class:`~alaric.comparison.GT`
    * :py:class:`~alaric.comparison.LT`
    * :py:class:`~alaric.comparison.EQ`


    Lets get all documents *without* a field called ``prefix``

    .. code-block:: python
        :linenos:

        from alaric.comparison import Exists
        from alaric.meta import Negate
        from alaric import AQ

        query = AQ(Negate(Exists("prefix")))
    """

    def __init__(self, comparison: ComparisonT | Exists):
        self.comparison: ComparisonT = comparison

    def __repr__(self):
        return f"NEGATE({self.comparison})"

    def build(self) -> Dict:
        """Returns a mongo usable filter for the negated option."""
        if isinstance(self.comparison, Exists):
            self.comparison._val = False
            return self.comparison.build()

        elif isinstance(self.comparison, IN):
            self.comparison._operator = "$nin"
            return self.comparison.build()

        elif isinstance(self.comparison, GT):
            log.debug("Use LT rather then negating GT")
            self.comparison = LT(self.comparison.field, self.comparison.value)

        elif isinstance(self.comparison, LT):
            log.debug("Use GT rather then negating LT")
            self.comparison = GT(self.comparison.field, self.comparison.value)

        elif isinstance(self.comparison, EQ):
            self.comparison._operator = "$ne"
            return self.comparison.build()

        raise RuntimeError("Invalid wrapped comparison.")



from __future__ import annotations

import hashlib
from typing import Union, Dict

from alaric import util
from alaric.abc import ComparisonT
from alaric.types import ObjectId


class HashedQueryField:
    """Use this to query against hashed fields.

    This class exposes an alias `HQF` for shorter usage.

    .. code-block:: python

        from alaric import AQ
        from alaric.comparison import EQ
        from alaric.encryption import HQF

        query = AQ(HQF(EQ("_id", 1)))
    """

    def __init__(self, entry: ComparisonT):
        self._entry: ComparisonT = entry

    def build(
        self,
    ) -> Dict[str, Dict[str, Union[int, str, float, bytes, dict, ObjectId]]]:
        """Return this instance as a usable Mongo filter."""
        initial = self._entry.build()
        out = {}
        for k, v in initial.items():
            d = {}
            for nested_k, nested_v in v.items():
                d[nested_k] = util.hash_field(nested_k, nested_v)
            out[k] = d

        return out



from typing import runtime_checkable, Protocol, Dict


@runtime_checkable
class Saveable(Protocol):
    """Protocol for class based Queries."""

    def as_dict(self) -> Dict:
        """Returns this class represented as a dictionary."""
        ...



from alaric.encryption import Base


class HashedFields(Base):
    """A list of fields which should be hashed when encountered.

    Due to the nature of hashing, this is a one way operation
    and is only done when sending items to Mongo.

    .. note::

        Take care not to pass the hashed output back
        to a method which sends data to Mongo as it will
        result in it being re-hashed.

        If this is the case, tell the document method to
        ignore processing the field.


    .. code-block:: python
        :linenos:

        from alaric.encryption import HashedFields

        HashedFields("guild_id", "test")
    """



from typing import Union, List, Dict

from alaric.abc import ComparisonT, LogicalT, Buildable


class AQ:
    """
    A container representing an advanced query.

    Parameters
    ----------
    item: Union[ComparisonT, LogicalT]]
        The parent item we wish to build upon.


    .. code-block:: python

        # A query to fetch all items
        # where the `id` field is equal to `1`
        # AND the document contains a `prefix` field
        from alaric import AQ
        from alaric.logical import AND
        from alaric.comparison import EQ, EXISTS

        query = AQ(AND(EQ("id", 1), EXISTS("prefix")))
    """

    def __init__(
        self,
        item: Union[ComparisonT, LogicalT, Buildable],
    ):
        self._item: Union[ComparisonT, LogicalT] = item

    def build(self) -> Dict:
        """Return this AQ as a usable Mongo filter."""
        return self._item.build()
